/*
 * rdf2dot.c
 * Turn a serel state xml file into a .dot AT+T GraphViz file,
 * after applying various transformations.
 *
 * Copyright (C) Leni Mayo 2002
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License, version 2, as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h> /* for malloc/free */
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>

#include "config.h"  /* generated by autoconf configure */
#include "traverse.h"
#include "rdfconsturi.h"
#include "rdfcommon.h"

int initialise(librdf_uri **, librdf_storage **,
		librdf_model **, char *pszSourceUri);

void cleanup(librdf_uri *pUri, librdf_storage *pStorage, librdf_model *pModel);

int parseUriIntoModel(librdf_model *pModel, librdf_uri *pUri);

void parseCommandLine(int argc, char *argv[], int *pOptind, GSList **ppListDotProperties, int *pfPruneEdges, int *pSlowest);
void printUsage(FILE *fp, const char *argv0);

int   rdf2dot(librdf_model *pModel, librdf_node *pGraph,
		GSList *pListDotProperties, int fPruneEdges, int slowest);
float rdf2dotWeightFromPath(librdf_model *pModel, GSList *pList);
int   rdf2dot_pre(librdf_model *pModel, librdf_node *pNode, DftUserData *);
int   mark_non_critical_edge_pre(librdf_model *pModel, librdf_node *pNode, DftUserData *);
int   mark_slowest_paths_pre(librdf_model *pModel, librdf_node *pNode, DftUserData *pUserData);

static void markCriticalPath(librdf_model *pModel, GSList *pRootNodes, GSList **ppCriticalPath);
static void markSlowestPaths(librdf_model *pModel, GSList *pRootNodes, GSList *pLeafNodes, int cutoff);
static void markNonCriticalEdges(librdf_model *pModel, librdf_node *pGraph);

static void printDotHeader(FILE *fp, GSList *pListProperties);
static void printDotTrailer(FILE *fp);
static void printDotStartPairs(FILE *fp, librdf_model *pModel, GSList *pRootNodes, GSList *pCriticalPath, int fOnlySlowestPaths);
static void printDotEdges(FILE *fp, librdf_model *pModel, librdf_node *pContainer, GSList *pCriticalPath, int fOnlySlowestPaths);
static void printDotNodes(FILE *fp, librdf_model *pModel, librdf_node *pContainer, GSList *pCriticalPath, int fOnlySlowestPaths);
static void printDotEdge(FILE *fp, const char *pszFrom, const char *pszTo, const char *pszProperty);

static int tagNodeWithRootToLeafWeight(librdf_model *pModel, librdf_node *pNode, float max);
static int tagNodeWithOnSlowestPath(librdf_model *pModel, librdf_node *pNode);
static int tagPathWithOnSlowestPath(librdf_model *pModel, GSList *pList);
static int tagPathWithRankSlowest(librdf_model *pModel, GSList *pLeafNodes);

static char * getDotEdgeDecoration(char buf[], int buflen, librdf_model *pModel, librdf_node *pFrom);
static char * rankToString(librdf_model *pModel, librdf_node *pSubject, char buf[], int buflen);

int countProperty(librdf_model *pModel, librdf_node *pNode, RdfProperty name);
int weightFromNode(librdf_model *pModel, librdf_node *pNode, RdfProperty name, float *pWeight);

int weightCompare(float a, float b);

typedef struct
{
	GSList *pList; /* path with the max weight */
	float  max;    /* the weight */
} MarkCriticalPathUserData;

typedef struct
{
	/*
	 * pNode is the "to" node, if we meet it in a traversal from the "from"
	 * node then the edge can't be on the critical path.
	 * found is set to TRUE of pNode was found, otherwise FALSE.
	 */
	librdf_node *pNode;
	int found;
} MarkNonCriticalEdgeUserData;

typedef struct
{
	GSList *pLeafList;
} MarkRootToLeafUserData;

static const char *pszCriticalEdgeColour = "red";
static const char *pszParserName = "repat";
static const char *pszDefaultArgv0 = "serel-rdf2dot";

gint rdf_node_compare_fn(gconstpointer a, gconstpointer b);
gint root_to_leaf_weight_compare_fn(gconstpointer a, gconstpointer b);
/*
 * leni TODO: Fix this a temporary hack.
 * root_to_leaf_weight_compare_fn needs a (librdf_model *) in order to 
 * do a comparison, but there's no way of getting it into the routine
 * other than a static or global.
 * I should probably make pModel a global, but just at the mo,
 * use this rotten hack.
 * leni - Mon Mar 25 14:17:20 EST 2002
 */
static librdf_model *pModelFor_root_to_leaf_weight_compare_fn;

int main(int argc, char *argv[]) 
{
	librdf_storage *pStorage;
	librdf_model *pModel;
	librdf_node *pGraph;
	librdf_uri *pSourceUri;
	int nPathArgc;
	GSList *pListDotProperties;
	int fPruneEdges;
	int slowest;
	int ret;

	parseCommandLine(argc, argv, &nPathArgc, &pListDotProperties, &fPruneEdges, &slowest);

	if (nPathArgc + 1 != argc)
	{
		printUsage(stderr, argv ? argv[0] : pszDefaultArgv0);
		return 1;
	}

#ifdef TAKENOUT
	fprintf(stderr, "After parsing command line: nPathArgc == %d, pListDotProperties[0] = %s, fPruneEdges == %d, slowest = %d\n",
		nPathArgc,
		(pListDotProperties == NULL ? "empty" : (char *) pListDotProperties->data),
		fPruneEdges,
		slowest);
#endif

	{
		char *pszUri = newUriStringEnsuringScheme(argv[nPathArgc]);

		ret = initialise(&pSourceUri, &pStorage, &pModel, pszUri);

		free(pszUri);
	}

	if (!ret)
	{
		fprintf(stderr, "Unable to initialise ... exiting.\n");

		return 1;
	}

	ret = ret && parseUriIntoModel(pModel, pSourceUri);

	if (!ret)
	{
		return 1;
	}

	pModelFor_root_to_leaf_weight_compare_fn = pModel;
	
#ifdef TAKENOUT
	printf("Before tagging, the model is:\n");
	librdf_model_print(pModel, stdout);
	fflush(stdout);
#endif

	/* Get a graph - don't handle multiples (or none) at the mo
	*/
	pGraph = librdf_model_get_source(pModel, nodeFromCode(eRdfType), nodeFromCode(eGraphGraph));

	if (pGraph)
	{
		ret = ret && rdf2dot(pModel, pGraph, pListDotProperties, fPruneEdges, slowest);
	}
	else
	{
		fprintf(stderr, "error: xml/rdf does not appear to contain a '%s' element.\n", rdfPropertyFragment(eGraphGraph));
	}

	cleanup(pSourceUri, pStorage, pModel);

	if (pListDotProperties != NULL)
	{
		GSList *pElement;

		for (pElement = pListDotProperties; pElement != NULL; pElement = pElement->next)
		{
			free(pElement->data);
		}

		g_slist_free(pListDotProperties);
	}

#ifdef LIBRDF_MEMORY_DEBUG
	librdf_memory_report(stderr);
#endif
	
	return 0;
}

/*
 * initialise()
 *
 * Return TRUE if all's well, false otherwise.
 */
int initialise(librdf_uri **ppUri, librdf_storage **ppStorage, librdf_model **ppModel, char *pszSourceUri)
{
	int ret;

	assert(ppUri && ppStorage && ppModel);

	pWorld = librdf_new_world();

	ret = pWorld != NULL;

	if (ret)
	{
		librdf_world_open(pWorld);

		*ppUri = librdf_new_uri(pWorld, pszSourceUri);

		ret = *ppUri != NULL;
	}

	if (ret)
	{
		*ppStorage = librdf_new_storage(pWorld, "hashes", "test", "hash-type='memory'");

		ret = *ppStorage != NULL;
	}

	if (ret)
	{
		*ppModel = librdf_new_model(pWorld, *ppStorage, NULL);

		ret = *ppModel != NULL;
	}

	ret = ret && rdfConstInitialise();

	return ret;
}

void cleanup(librdf_uri *pUri, librdf_storage *pStorage, librdf_model *pModel)
{
	rdfConstCleanup();

	librdf_free_model(pModel);
	librdf_free_storage(pStorage);
	librdf_free_uri(pUri);
	librdf_free_world(pWorld);
}

/*
 * parseUriIntoModel()
 *
 * Return TRUE if all's well, false otherwise.
 */
int parseUriIntoModel(librdf_model *pModel, librdf_uri *pUri)
{
	librdf_parser *pParser;
	int ret;

	pParser = librdf_new_parser(pWorld, pszParserName, NULL, NULL);

	ret = pParser != NULL;

	if (!ret)
	{
		fprintf(stderr, "Unable to create parser: %s\n", pszParserName);

		return ret;
	}

	ret = librdf_parser_parse_into_model(pParser, pUri, pUri, pModel) == 0;

	if (!ret)
	{
		fprintf(stderr, "Failed to parse rdf into model\n");
	}

	librdf_free_parser(pParser);

	return ret;
}

#ifdef TAKENOUT
/*
 * Test assumptions about the parsed rdf document that we rely on later:
 *	- Graph statement has properties: nodes and edges
 *	- nodes property references all node statements (incl no dangling node statements)
 *	- edges property references all edge statements (incl no dangling edge statements)
 *
 * If this test fails, it's not that the graph has bad syntax or is otherwise an
 * invalid graph, it's that we can't use the the document.
 *
 * Not implemented!
 */
int testAssumptions(librdf_model *pModel, librdf_node *pGraph)
{
	librdf_node *pContainerNode;
	librdf_node *pContainerEdge;
	int ret = TRUE;

	pContainerNode = librdf_model_get_target(pModel, pGraph, nodeFromCode(eGraphNodes));

	pContainerEdge = librdf_model_get_target(pModel, pGraph, nodeFromCode(eGraphEdges));


	return ret;
}
#endif

/*
 * int rdf2dot();
 *
 * Returns:
 *  TRUE iff
 *	- the model is acyclic
 *  - the model was succesfully output as a AT+T GraphViz dot file.
 */
int rdf2dot(librdf_model *pModel, librdf_node *pGraph, GSList *pListDotProperties, int fPruneEdges, int slowest)
{
	int ret = TRUE;
	GSList *pRootNodes, *pLeafNodes;
	librdf_node *pContainerNode;
	librdf_node *pContainerEdge;
	GSList *pCycle, *pCriticalPath;

	assert(pGraph);

#ifdef TAKENOUT
	ret = testAssumptions(pModel, pGraph);
#endif

	pContainerNode = librdf_model_get_target(pModel, pGraph, nodeFromCode(eGraphNodes));

	pContainerEdge = librdf_model_get_target(pModel, pGraph, nodeFromCode(eGraphEdges));

	pRootNodes = listOfReferencedNode(pModel, nodeFromCode(eGraphNode), nodeFromCode(eGraphTo));
	pLeafNodes = listOfReferencedNode(pModel, nodeFromCode(eGraphNode), nodeFromCode(eGraphFrom));

#ifdef TAKENOUT
	pRootNodes = nodes_from_rdf_container(pModel, pContainerNode, nodeFromCode(eGraphTo));
	pLeafNodes = nodes_from_rdf_container(pModel, pContainerNode, nodeFromCode(eGraphFrom));

	printf("In rdf2dot():\n");
	printf("pRootNodes = ");
	print_nodes_in_list(stdout, pRootNodes);
	printf("\n");
	printf("pLeafNodes = ");
	print_nodes_in_list(stdout, pLeafNodes);
	printf("\n");
#endif

	testAcyclic(pModel, &pCycle, nodeFromCode(eGraphNode));

	ret = (pCycle == NULL);

	if (pCycle != NULL)
	{
		char *psz = newStringFromNodesInList(pCycle);

		fprintf(stderr, "Error: graph contains a cycle: %s\n", psz);

		free(psz);

		return ret;
	}

	markCriticalPath(pModel, pRootNodes, &pCriticalPath);

	if (pCriticalPath == NULL || pRootNodes == NULL || pLeafNodes == NULL)
	{
		fprintf(stderr, "Graph appears to be empty\n");

		return ret;
	}

	pLeafNodes = g_slist_sort(pLeafNodes, root_to_leaf_weight_compare_fn);
	pLeafNodes = g_slist_reverse(pLeafNodes);

	tagPathWithRankSlowest(pModel, pLeafNodes);

	if (fPruneEdges)
	{
		markNonCriticalEdges(pModel, pGraph);
	}

	if (slowest != 0)
	{
		markSlowestPaths(pModel, pRootNodes, pLeafNodes, slowest);
	}

#ifdef TAKENOUT
	printf("After tagging, the model is:\n");
	librdf_model_print(pModel, stdout);
	fflush(stdout);
#endif

	printDotHeader(stdout, pListDotProperties);

	printDotStartPairs(stdout, pModel, pRootNodes, pCriticalPath, slowest != 0);

	printDotEdges(stdout, pModel, pContainerEdge, pCriticalPath, slowest != 0);

	printDotNodes(stdout, pModel, pContainerNode, pCriticalPath, slowest != 0);

	printDotTrailer(stdout);

	g_slist_free(pRootNodes);
	g_slist_free(pLeafNodes);

	return ret;
}

void markCriticalPath(librdf_model *pModel, GSList *pRootNodes, GSList **ppCriticalPath)
{
	DftUserData user_data;
	MarkCriticalPathUserData ud;

	ud.pList = NULL;

	user_data.pData = &ud;
	
	depth_first_traverse_start(pModel, pRootNodes, nodeFromCode(eGraphFrom), nodeFromCode(eGraphTo), rdf2dot_pre, stack_pop_callback_post, &user_data);

	if (ud.pList)
	{
		ud.pList = g_slist_reverse(ud.pList);
	}

	assert(user_data.pStack == NULL); /* cycles are handled elsewhere */
	*ppCriticalPath = ud.pList;

#ifdef TAKENOUT
	printf("markCriticalPath() returns: ");
	print_nodes_in_list(stdout, *ppCriticalPath);
#endif
}

void markSlowestPaths(librdf_model *pModel, GSList *pRootNodes, GSList *pLeafNodes, int cutoff)
{
	DftUserData user_data;
	MarkRootToLeafUserData ud;
	GSList *pElement;
	int count = 0;

	ud.pLeafList = NULL;

	user_data.pData = &ud;
	
	for (pElement = pLeafNodes; pElement != NULL && count < cutoff; pElement = pElement->next)
	{
		assert(pElement->data);
		count++;

		ud.pLeafList = g_slist_prepend(ud.pLeafList, pElement->data);
	}

#ifdef TAKENOUT
	fprintf(stderr, "markSlowestPaths looking to mark paths to these leaves: ");
	print_nodes_in_list(stderr, ud.pLeafList);
	fprintf(stderr, "\n");
#endif

	depth_first_traverse_start(pModel, pRootNodes, nodeFromCode(eGraphFrom), nodeFromCode(eGraphTo), mark_slowest_paths_pre, stack_pop_callback_post, &user_data);

	g_slist_free(ud.pLeafList);
}

void markNonCriticalEdges(librdf_model *pModel, librdf_node *pGraph)
{
	DftUserData user_data;
	librdf_node *pContainerEdge;
	librdf_node *pNode, *pFrom, *pTo;
	GSList *pList = NULL;
	int count = 0;
	MarkNonCriticalEdgeUserData ud;
	int ret;

	user_data.pData = &ud;

	pContainerEdge = librdf_model_get_target(pModel, pGraph, nodeFromCode(eGraphEdges));

	pList = g_slist_append(pList, NULL); /* list has one element */

	while (TRUE)
	{
		count++;

		pNode = nodeAtContainerItem(pModel, pContainerEdge, count);

		if (pNode != NULL)
		{
			pFrom = librdf_model_get_target(pModel, pNode, nodeFromCode(eGraphFrom));
			pTo = librdf_model_get_target(pModel, pNode, nodeFromCode(eGraphTo));

			if (!pFrom || !pTo)
			{
				/* Something is very screwy about this edge, just ignore it
				*/
				continue;
			}

			if (countProperty(pModel, pTo, eGraphTo) < 2)
			{
				/* this is the only edge that points to this node
				*/
				continue;
			}

			ud.pNode = pTo;
			
			pList->data = pFrom;

#ifdef TAKENOUT
			printf("markNonCriticalEdges() doing dft from %s looking for %s\n",
				local_part_of_uri(node_as_uri_string(pFrom)),
				local_part_of_uri(node_as_uri_string(pTo)) );
#endif

			ret = addPropertyToNode(pModel, pNode, eSerelMarkDeleted, "deleted");

			if (!ret)
			{
				fprintf(stderr, "markNonCriticalEdges(), addPropertyToNode(%s, %s, %s) failed!\n",
					local_part_of_uri(node_as_uri_string(pNode)),
					local_part_of_uri(node_as_uri_string(nodeFromCode(eSerelMarkDeleted))), 
					"failed" );
			}

			depth_first_traverse_start(pModel, pList, nodeFromCode(eGraphFrom),
				nodeFromCode(eGraphTo), mark_non_critical_edge_pre, stack_pop_callback_post, &user_data);

			ret = removeStatement(pModel, pNode, nodeFromCode(eSerelMarkDeleted), NULL);

			if (!ret)
			{
				fprintf(stderr, "markNonCriticalEdges(), removeStatement(%s, %s, NULL) failed!\n",
					local_part_of_uri(node_as_uri_string(pNode)),
					local_part_of_uri(node_as_uri_string(nodeFromCode(eSerelMarkDeleted))) );
			}

			if (ud.found)
			{
				addPropertyToNode(pModel, pNode, eSerelMarkDeleted, "deleted");
				/* printf("markNonCriticalEdges() marks %s as deleted\n", local_part_of_uri(node_as_uri_string(pNode)) ); */
			}
		}
		else
		{
			count--;
			break;
		}
	}

	g_slist_free(pList);
}

/*
 * int rdf2dot_pre(librdf_model *pModel, librdf_node *pNode, void *pUserData)
 *
 * Callback routine used to determine the critical path.
 */

int rdf2dot_pre(librdf_model *pModel, librdf_node *pNode, DftUserData *pUserData)
{
	int ret;
	MarkCriticalPathUserData *pUd;
	float max;

	assert(pModel && pNode && pUserData);

	pUd = (MarkCriticalPathUserData *) pUserData->pData;

	ret = stack_push_callback_pre(pModel, pNode, pUserData);

	/* only do a calculation if pNode is a leaf
	*/
	if (ret && !librdf_model_has_arc_in(pModel, pNode, nodeFromCode(eGraphFrom)))
	{
		max = rdf2dotWeightFromPath(pModel, pUserData->pStack);

		tagNodeWithRootToLeafWeight(pModel, pNode, max);

#ifdef TAKENOUT
		printf("rdf2dot_pre(%s) calculates %4.2f as weight (%s) for path: ",
			local_part_of_uri(node_as_uri_string(pNode)),
			max,
			((max > pUd->max) ? "largest" : "not largest") );

		print_nodes_in_list(stdout, pUserData->pStack);
#endif

		if (max > pUd->max)
		{
			g_slist_free(pUd->pList);

			pUd->max = max;
			
			pUd->pList = g_slist_copy(pUserData->pStack);
		}

#ifdef TAKENOUT
		printf("Critical path at this point is: %4.2f: ", pUd->max);
		print_nodes_in_list(stdout, pUd->pList);
		printf("\n");
#endif

	}

	return ret;
}

int mark_slowest_paths_pre(librdf_model *pModel, librdf_node *pNode, DftUserData *pUserData)
{
	int ret;
	MarkRootToLeafUserData *pUd;
	float weightOfPath, weightLeaf;
	int fPresent;

	assert(pModel && pNode && pUserData);

	pUd = (MarkRootToLeafUserData *) pUserData->pData;

	ret = stack_push_callback_pre(pModel, pNode, pUserData);

	/* only interested if pNode is a leaf
	*/
	if (ret &&
	    !librdf_model_has_arc_in(pModel, pNode, nodeFromCode(eGraphFrom)) &&
		g_slist_find_custom(pUd->pLeafList, pNode, rdf_node_compare_fn))
	{
		weightOfPath = rdf2dotWeightFromPath(pModel, pUserData->pStack);

		fPresent = weightFromNode(pModel, pNode, eSerelRootToLeafWeight, &weightLeaf);

		if (fPresent && weightCompare(weightOfPath, weightLeaf) == 0)
		{
			tagPathWithOnSlowestPath(pModel, pUserData->pStack);
		}
	}

	return ret;
}

int mark_non_critical_edge_pre(librdf_model *pModel, librdf_node *pNode, DftUserData *pUserData)
{
	int fCycle;
	MarkNonCriticalEdgeUserData *pUd;

	assert(pModel && pNode && pUserData);

	pUd = (MarkNonCriticalEdgeUserData *) pUserData->pData;

	fCycle = ! stack_push_callback_pre(pModel, pNode, pUserData);

	assert(!fCycle);  /* cycles should have been detected and dealt with elsewhere */

	pUd->found = librdf_node_equals(pNode, pUd->pNode) != 0;

#ifdef TAKENOUT
	printf("mark_non_critical_edge_pre() visits %s, compares it with %s, says %s\n",
		local_part_of_uri(node_as_uri_string(pNode)),
		local_part_of_uri(node_as_uri_string(pUd->pNode)),
		pUd->found ? "match" : "don't match" );
#endif

	return ! pUd->found;
}


float rdf2dotWeightFromPath(librdf_model *pModel, GSList *pList)
{
	float total = 0;
	float weight;
	GSList *pElement;
	int found = TRUE;

	for (pElement = pList; pElement != NULL && found; pElement = pElement->next)
	{
		assert(pElement->data);

		found = weightFromNode(pModel, (librdf_node *) pElement->data, eGraphWeight, &weight);
		
		total += weight;
	}

	if (!found)
	{
		total = 0;
	}

	return total;
}

void printDotHeader(FILE *fp, GSList *pListProperties)
{
	char bufTime[100];
	time_t t = time(NULL);
	struct tm *pTm = gmtime(&t);
	const char *pszProperty;
	GSList *pList = g_slist_copy(pListProperties);
	GSList *pElement;
	char buf[200];
	int i;
	struct pair
	{
		const char *pszKeyword;
		const char *pszValue;
	};
	/* leni TODO: probably want to remove these fontname defaults for general distribution.
	*/
	struct pair table[] =
	{
		{ "node",    "[fontname=\"/usr/share/fonts/default/Type1/n019023l.pfb\", fontsize=10, shape=ellipse];" },
		{ "edge",    "[fontname=\"/usr/share/fonts/default/Type1/n019023l.pfb\", fontsize=11];" },
		{ "ratio",   " = auto;" },
		{ "rankdir", "= LR;" },
		/* { "size",    " = \"7.5,10\";" }, */
		{ NULL, NULL }
	};

	strftime(bufTime, sizeof(bufTime), "%G-%m-%d %T UTC", pTm);

	fprintf(fp, "digraph G {\n");
	fprintf(fp, "comment=\"Generated by serel on %s\";\n", bufTime);

	for (i = 0; table[i].pszKeyword != NULL; i++)
	{
		pszProperty = NULL;

		if (pListProperties)
		{
			for (pElement = pList; pElement != NULL; pElement = pElement->next)
			{
				assert(pElement->data);

				if (strncmp(table[i].pszKeyword,
				       (char *) pElement->data, strlen(table[i].pszKeyword)) == 0)
				{
					pszProperty = pElement->data;
					pList = g_slist_remove_link(pList, pElement);
					g_slist_free_1(pElement);
					break;
				}
			}
		}

		if (pszProperty == NULL)
		{
			snprintf(buf, sizeof(buf), "%s %s", table[i].pszKeyword, table[i].pszValue);
			pszProperty = buf;
		}

		fprintf(fp, "%s\n", pszProperty);
	}

	for (pElement = pList; pElement != NULL; pElement = pElement->next)
	{
		assert(pElement->data);

		fprintf(fp, "%s\n", (char *) pElement->data);
	}

	fprintf(fp, "\n");

	fprintf(fp, "start [shape=plaintext, label=\"start\"];\n");
	fprintf(fp, "\n");
}

void printDotTrailer(FILE *fp)
{
	fprintf(fp, "}\n");
}

void printDotStartPairs(FILE *fp, librdf_model *pModel, GSList *pList, GSList *pCriticalPath, int fOnlySlowestPaths)
{
	GSList *pElement;
	const char *pszProperty;
	static char buf[100];
	librdf_node *pNode;

	for (pElement = pList; pElement != NULL; pElement = pElement->next)
	{
		assert(pElement->data);

		pNode = (librdf_node *) pElement->data;

		if (fOnlySlowestPaths && ! nodeHasPropertyName(pModel, pNode, eSerelOnSlowestPath))
		{
			continue;
		}

		if (g_slist_find_custom(pCriticalPath, pNode, rdf_node_compare_fn))
		{
			snprintf(buf, sizeof(buf), "[color = \"%s\"]", pszCriticalEdgeColour);
			pszProperty = buf;
		}
		else
		{
			pszProperty = NULL;
		}

		printDotEdge(fp, "start", local_part_of_uri(node_as_uri_string(pNode)), pszProperty);
	}

	fprintf(fp, "\n");
}

void printDotNodes(FILE *fp, librdf_model *pModel, librdf_node *pContainer, GSList *pCriticalPath, int fOnlySlowestPaths)
{
	librdf_node *pNode;
	int count = 0;
	const char *pszProperty;
	char bufProperty[100];
	char bufFinish[100];
	char bufRank[10];
	int fIsLeaf;
	float weight;

	while (TRUE)
	{
		count++;

		pNode = nodeAtContainerItem(pModel, pContainer, count);

		if (pNode)
		{
			fIsLeaf = weightFromNode(pModel, pNode, eSerelRootToLeafWeight, &weight);

			if (fIsLeaf)
			{
				if (fOnlySlowestPaths && ! nodeHasPropertyName(pModel, pNode, eSerelOnSlowestPath))
				{
					continue;
				}

				if (g_slist_find_custom(pCriticalPath, pNode, rdf_node_compare_fn))
				{
					pszProperty = getDotEdgeDecoration(bufProperty, sizeof(bufProperty), pModel, pNode);
				}
				else
				{
					pszProperty = NULL;
				}

				snprintf(bufFinish, sizeof(bufFinish), "serel_finish_%s", local_part_of_uri(node_as_uri_string(pNode)));

				fprintf(fp, "%s [shape=plaintext, label=\"%s (%4.2f s)\"];\n",
					bufFinish,
					rankToString(pModel, pNode, bufRank, sizeof(bufRank)),
					weight);

				printDotEdge(fp, local_part_of_uri(node_as_uri_string(pNode)), bufFinish, pszProperty);
			}
		}
		else
		{
			count--;
			break;
		}
	}
}

void printDotEdges(FILE *fp, librdf_model *pModel, librdf_node *pContainer, GSList *pCriticalPath, int fOnlySlowestPaths)
{
	librdf_node *pNode;
	librdf_node *pFrom, *pTo;
	GSList *pList;
	int count = 0;
	const char *pszProperty;
	char buf[100];

	while (TRUE)
	{
		count++;

		pNode = nodeAtContainerItem(pModel, pContainer, count);

		if (pNode != NULL)
		{
			if (nodeHasPropertyName(pModel, pNode, eSerelMarkDeleted))
			{
				continue;
			}

			pFrom = librdf_model_get_target(pModel, pNode, nodeFromCode(eGraphFrom));
			pTo = librdf_model_get_target(pModel, pNode, nodeFromCode(eGraphTo));

			if (!pFrom || !pTo)
			{
				/* something screwy with this edge
				*/
				continue;
			}

			if (fOnlySlowestPaths &&
			    ( ! nodeHasPropertyName(pModel, pFrom, eSerelOnSlowestPath) ||
			      ! nodeHasPropertyName(pModel, pTo, eSerelOnSlowestPath)))
			{
				continue;
			}

			/*
			 * Note that the test for whether and edge is on the critical
			 * path relies on the root-to-leaf ordering of pCriticalPath
			 * This is guaranteed by the nature of the depth-first-traversal.
			 */
			pList = g_slist_find_custom(pCriticalPath, pFrom, rdf_node_compare_fn);

			if (pList && pList->next && !rdf_node_compare_fn(pList->next->data, pTo))
			{
				pszProperty = getDotEdgeDecoration(buf, sizeof(buf), pModel, pFrom);
			}
			else
			{
				pszProperty = NULL;
			}

			printDotEdge(fp,
				local_part_of_uri(node_as_uri_string(pFrom)),
				local_part_of_uri(node_as_uri_string(pTo)),
				pszProperty);
		}
		else
		{
			count--;
			break;
		}
	}
}

void printDotEdge(FILE *fp, const char *pszFrom, const char *pszTo, const char *pszProperty)
{
	assert(pszFrom && pszTo);

	fprintf(fp, "%s -> %s", pszFrom, pszTo);

	if (pszProperty)
	{
		fprintf(fp, " %s", pszProperty);
	}

	fprintf(fp, ";\n");
}

gint rdf_node_compare_fn(gconstpointer a, gconstpointer b)
{
	return !librdf_node_equals((librdf_node *) a, (librdf_node *) b);
}


gint root_to_leaf_weight_compare_fn(gconstpointer a, gconstpointer b)
{
	int fA, fB;
	float weightA, weightB;
	int ret = 0;
	librdf_model *pModel = pModelFor_root_to_leaf_weight_compare_fn;

	assert(a && b);

	fA = weightFromNode(pModel, (librdf_node *) a, eSerelRootToLeafWeight, &weightA);

	fB = weightFromNode(pModel, (librdf_node *) b, eSerelRootToLeafWeight, &weightB);

	if (fA && fB)
	{
		ret = weightCompare(weightA, weightB);
	}

	return ret;
}

char * getDotEdgeDecoration(char buf[], int buflen, librdf_model *pModel, librdf_node *pFrom)
{
	librdf_node *pTime;
	const char *pszTime = NULL;

	pTime = librdf_model_get_target(pModel, pFrom, nodeFromCode(eGraphWeight));

	if (pTime)
	{
		pszTime = librdf_node_get_literal_value(pTime);
	}

	if (pszTime && strlen(pszTime) < 20)
	{
		snprintf(buf, buflen, "[color = \"%s\", label = \"%s s\"]", pszCriticalEdgeColour, pszTime);
	}

	return buf;
}

int tagNodeWithRootToLeafWeight(librdf_model *pModel, librdf_node *pNode, float max)
{
	float weight = 0;
	int   fPresent;
	char bufWeight[100];
	int ret = TRUE;
	int tagged = FALSE;

	fPresent = weightFromNode(pModel, pNode, eSerelRootToLeafWeight, &weight);

	if (!fPresent || max > weight)
	{
		tagged = TRUE;

		sprintf(bufWeight, "%4.2f", max);

		if (fPresent)
		{
			ret = removeStatement(pModel, pNode, nodeFromCode(eSerelRootToLeafWeight), NULL);
		}

		ret = ret && addPropertyToNode(pModel, pNode, eSerelRootToLeafWeight, bufWeight);
	}

#ifdef TAKENOUT
	printf("tagNodeWithRootToLeafWeight(%s, %4.2f) (%s) ... %s\n",
			local_part_of_uri(node_as_uri_string(pNode)),
			max,
			(fPresent ? "    present" : "not present"),
			(tagged ? "tagged" : "skipped"));
	
	fPresent = weightFromNode(pModel, pNode, eSerelRootToLeafWeight, &weight);

	assert(fPresent);

	printf("After which, weightFromNode() returns %4.2f\n", weight);
#endif

	return ret;
}

int tagNodeWithOnSlowestPath(librdf_model *pModel, librdf_node *pSubject)
{
	librdf_node *pObject;
	int ret = TRUE;
	int tagged = FALSE;

	pObject = librdf_model_get_target(pModel, pSubject, nodeFromCode(eSerelOnSlowestPath));

	if (pObject == NULL)
	{
		ret = addPropertyToNode(pModel, pSubject, eSerelOnSlowestPath, "present");

		tagged = TRUE;
	}

#ifdef TAKENOUT
	fprintf(stdout, "tagNodeWithOnSlowestPath(%s) ... %s\n",
		local_part_of_uri(node_as_uri_string(pSubject)),
		(tagged ? "tagged" : "skipped"));
#endif

	return ret;
}

int tagPathWithOnSlowestPath(librdf_model *pModel, GSList *pList)
{
	int ret = TRUE;
	GSList *pElement;

	for (pElement = pList; pElement != NULL && ret; pElement = pElement->next)
	{
		assert(pElement->data);
		
		ret = tagNodeWithOnSlowestPath(pModel, (librdf_node *) pElement->data);
	}

	return ret;
}

int tagPathWithRankSlowest(librdf_model *pModel, GSList *pList)
{
	int ret = TRUE;
	GSList *pElement;
	librdf_node *pNode;
	char buf[10];
	int count = 0;

	for (pElement = pList; pElement != NULL && ret; pElement = pElement->next)
	{
		assert(pElement->data);
		count++;

		pNode = pElement->data;

		snprintf(buf, sizeof(buf), "%d", count);
		
		ret = addPropertyToNode(pModel, pNode, eSerelRankSlowest, buf);
	}

	return ret;
}

int weightFromNode(librdf_model *pModel, librdf_node *pNode, RdfProperty name, float *pWeight)
{
	const char *pszWeight;
	librdf_node *pObject;
	int ret;

	assert(pWeight);

	pObject = librdf_model_get_target(pModel, pNode, nodeFromCode(name));

	ret = pObject != NULL;

	if (ret)
	{
		pszWeight = librdf_node_get_literal_value(pObject);

		ret = pszWeight != NULL;

		if (ret)
		{
			*pWeight = atof(pszWeight);
		}
	}

	return ret;
}

/*
 * int countProperty(librdf_model *pModel, librdf_node *pNode, RdfProperty name)
 *
 * Count how many times a Node is the value of a given property.
 * eg. how many times a vertex appears as the value of the "to" property
 * gives us the in-degree of the vertex.
 */
int countProperty(librdf_model *pModel, librdf_node *pNode, RdfProperty name)
{
	librdf_iterator *iterator;
	librdf_node *subject;
	librdf_node *pProperty = nodeFromCode(name);
	int count = 0;

	assert(pNode);

	iterator = librdf_model_get_sources(pModel, pProperty, pNode);

#ifdef TAKENOUT
	printf("countProperty iterating across: ");
	print_nodes_in_iterator(pModel, iterator);
	fflush(stdout);
#endif

	while ( !librdf_iterator_end(iterator))
	{
		subject = (librdf_node *) librdf_iterator_get_next(iterator);
		
		if (! nodeHasPropertyName(pModel, subject, eSerelMarkDeleted))
		{
			count++;
		}

		librdf_free_node(subject);
	}

	librdf_free_iterator(iterator);

#ifdef TAKENOUT
	printf("countProperty(%s, %s) returns %d\n",
		local_part_of_uri(node_as_uri_string(pNode)),
		local_part_of_uri(node_as_uri_string(nodeFromCode(name))),
		count );
#endif

	return count;
}

char *rankToString(librdf_model *pModel, librdf_node *pSubject, char buf[], int buflen)
{
	librdf_node *pObject;
	char *pszRank;
	int num = 0;
	const char *pszSuffix;
	int mod;

	pObject = librdf_model_get_target(pModel, pSubject, nodeFromCode(eSerelRankSlowest));

	if (pObject != NULL)
	{
		pszRank = librdf_node_get_literal_value(pObject);

		if (pszRank)
		{
			num = atoi(pszRank);
		}
	}

	mod = num % 10;

	switch(mod)
	{
		case 1:  pszSuffix = "st"; break;
		case 2:  pszSuffix = "nd"; break;
		case 3:  pszSuffix = "rd"; break;
		default: pszSuffix = "th"; break;
	}

	if (num == 1)
	{
		snprintf(buf, buflen, "last");
	}
	else if (num < 4)
	{
		snprintf(buf, buflen, "%d%s last", num, pszSuffix);
	}
	else
	{
		snprintf(buf, buflen, "finish");
	}

	return buf;
}

int weightCompare(float a, float b)
{
	return 1000 * (a - b);
}


void parseCommandLine(int argc, char *argv[], int *pOptind, GSList **ppListDotProperties, int *pfPruneEdges, int *pSlowest)
{
	int c;
	char *p;
	int option_index = 0;
	static struct option long_options[] =
	{
		{"help",        no_argument,       NULL,  0 },
		{"version",     no_argument,       NULL,  0 },
		{"add",         required_argument, NULL, 'a'},
		{"prune-edges", no_argument,       NULL, 'e'},
		{"slowest",     required_argument, NULL, 's'},
		{0, 0, 0, 0}
	};

	assert(pOptind && ppListDotProperties && pfPruneEdges && pSlowest);

	*ppListDotProperties = NULL;
	*pfPruneEdges = FALSE;
	*pSlowest = 0;

	while ((c = getopt_long (argc, argv, "a:es:", long_options, &option_index)) != -1)
	{
		switch(c)
		{
		case 0:
			if (strcmp(long_options[option_index].name, "help") == 0)
			{
				printUsage(stdout, argv ? argv[0] : pszDefaultArgv0);
			}
			else if (strcmp(long_options[option_index].name, "version") == 0)
			{
				printVersion(stdout, PACKAGE, VERSION);
			}

			exit(0);
			break;
		case 'a':
			/* The list and it's items are freed by the caller
			*/
			p = malloc(strlen(optarg) + 1);
			strcpy(p, optarg);
			*ppListDotProperties = g_slist_append(*ppListDotProperties, p);
			break;
		case 'e':
			*pfPruneEdges = TRUE;
			break;
		case 's':
			*pSlowest = atoi(optarg);
			break;
		default:
			printUsage(stderr, argv ? argv[0] : pszDefaultArgv0);
			exit(1);
		}
	}

	*pOptind = optind;
}

void printUsage(FILE *fp, const char *argv0)
{
	assert(argv0 != NULL);

	fprintf(fp, "Usage: serel-rdf2dot [options] path \n");
	fprintf(fp, "   -e --prune-edges      Prune redundant edges from the tree.\n");
	fprintf(fp, "   -a --add attribute    Add attributes to the generated dot file.\n");
	fprintf(fp, "   -s --slowest n        Only show the n slowest root-to-leaf paths.\n");
}
